@@ -28,7 +28,6 @@
 # include "builtins.h"
 # include "class.h"
 
-# include <time.h>
 # include <assert.h>
 # include <string.h>
 # include <stdarg.h>
@@ -95,9 +94,12 @@ void backtrace_line( FRAME *frame );
 void print_source_line( PARSE* p );
 
 
+struct frame *frame_before_python_call;
+
 void frame_init( FRAME* frame )
 {
     frame->prev = 0;
+    frame->prev_user = 0;
     lol_init(frame->args);
     frame->module = root_module();
     frame->rulename = "module scope";
@@ -266,7 +268,7 @@ compile_foreach(
         
         if ( parse->num )
         {
-            s = addsettings( s, 0, parse->string, L0 );
+            s = addsettings( s, VAR_SET, parse->string, L0 );
             pushsettings( s );
         }
 
@@ -492,7 +494,7 @@ compile_local(
     /* Initial value is ns */
 
     for( l = nt; l; l = list_next( l ) )
-        s = addsettings( s, 0, l->string, list_copy( (LIST*)0, ns ) );
+        s = addsettings( s, VAR_SET, l->string, list_copy( (LIST*)0, ns ) );
 
     list_free( ns );
     list_free( nt );
@@ -584,6 +586,7 @@ compile_rule(
 
     frame_init( inner );
     inner->prev = frame;
+    inner->prev_user = frame->module->user_module ? frame : frame->prev_user;
     inner->module = frame->module; /* This gets fixed up in evaluate_rule(), below */
     inner->procedure = parse;
 
@@ -690,6 +693,7 @@ type_check( char* type_name, LIST *values, FRAME* caller, RULE* called, LIST* ar
         frame_init( frame );
         frame->module = typecheck;
         frame->prev = caller;
+        frame->prev_user = caller->module->user_module ? caller : caller->prev_user;
 
         enter_module( typecheck );
         /* Prepare the argument list */
@@ -781,7 +785,7 @@ collect_arguments( RULE* rule, FRAME* frame )
                         }
                     }
                 
-                    locals = addsettings( locals, 0, name, value );
+                    locals = addsettings( locals, VAR_SET, name, value );
                     type_check( type_name, value, frame, rule, arg_name );
                     type_name = 0;
                 }
@@ -796,95 +800,115 @@ collect_arguments( RULE* rule, FRAME* frame )
     return locals;
 }
 
-struct profile_info
+static int python_instance_number = 0;
+
+RULE *
+enter_rule( char *rulename, module_t *target_module );
+
+#ifdef HAVE_PYTHON
+static LIST*
+call_python_function(RULE* r, FRAME* frame)
 {
-    char* name;                 /* name of rule being called */
-    clock_t cumulative;         /* cumulative time spent in rule */
-    clock_t net;                /* time spent in rule proper */
-    unsigned long num_entries;  /* number of time rule was entered */
-    unsigned long stack_count;  /* number of the times this function is present in stack */
-};
-typedef struct profile_info profile_info;
+    LIST* result = 0;
+    PyObject* arguments = PyTuple_New(frame->args->count);
+    int i ;
+    PyObject* py_result;
 
-struct profile_frame
-{
-    profile_info* info;               /* permanent storage where data accumulates */
-    clock_t overhead;                 /* overhead for profiling in this call */
-    clock_t entry_time;               /* time of last entry to rule */
-    struct profile_frame* caller;     /* stack frame of caller */
-    clock_t subrules;                 /* time spent in subrules */
-};
-typedef struct profile_frame profile_frame;
+    for(i = 0; i < frame->args->count; ++i)
+    {
+        PyObject* arg = PyList_New(0);
+        LIST* l = lol_get( frame->args, i);
 
-static profile_frame* profile_stack = 0;
-static struct hash* profile_hash = 0;
+        for(; l; l = l->next)
+        {
+            PyObject* v = PyString_FromString(l->string);
+            /* Steals reference to 'v' */
+            PyList_Append(arg, v);            
+        }
+        /* Steals reference to 'arg' */
+        PyTuple_SetItem(arguments, i, arg);
+    }
 
-static void profile_enter( char* rulename, profile_frame* frame )
-{
-    clock_t start = clock();
-    profile_info info, *p = &info;
-    
-    if ( !profile_hash )
-        profile_hash = hashinit(sizeof(profile_info), "profile");
-
-    info.name = rulename;
-    
-    if ( hashenter( profile_hash, (HASHDATA **)&p ) )
-        p->cumulative = p->net = p->num_entries = p->stack_count = 0;
-
-    ++(p->num_entries);
-    ++(p->stack_count);
-    
-    frame->info = p;
-    
-    frame->caller = profile_stack;
-    profile_stack = frame;
-
-    frame->entry_time = clock();
-    frame->overhead = 0;
-    frame->subrules = 0;
-
-    /* caller pays for the time it takes to play with the hash table */
-    if ( frame->caller )
-        frame->caller->overhead += frame->entry_time - start;
-}
-    
-static void profile_exit(profile_frame* frame)
-{
-    /* cumulative time for this call */
-    clock_t t = clock() - frame->entry_time - frame->overhead;
-    /* If this rule is already present on the stack, don't add the time for
-       this instance. */
-    if (frame->info->stack_count == 1)
-        frame->info->cumulative += t;
-    /* Net time does not depend on presense of the same rule in call stack. */
-    frame->info->net += t - frame->subrules;
+    frame_before_python_call = frame;
+    py_result = PyObject_CallObject(r->python_function, arguments);
+    Py_DECREF(arguments);
+    if (py_result != NULL) {
         
-    if (frame->caller)
-    {
-        /* caller's cumulative time must account for this overhead */
-        frame->caller->overhead += frame->overhead;
-        frame->caller->subrules += t;
+        if (PyList_Check(py_result)) {
+            int size = PyList_Size(py_result);
+            int i;
+            for(i = 0; i < size; ++i)
+            {
+                PyObject* item = PyList_GetItem(py_result, i);
+                if (PyString_Check(item))
+                {
+                    result = list_new(result, 
+                                      newstr(PyString_AsString(item)));
+                }
+                else
+                {
+                    fprintf(stderr, "Non-string object returned by Python call\n");
+                }
+            }
+        }
+        else if (PyInstance_Check(py_result))
+        {
+            static char instance_name[1000];
+            static char imported_method_name[1000];
+            module_t* m;
+            PyObject* method;
+            PyObject* method_name = PyString_FromString("foo");
+            RULE* r;
+
+            fprintf(stderr, "Got instance!\n");
+
+            snprintf(instance_name, 1000,
+                     "pyinstance%d", python_instance_number);
+            snprintf(imported_method_name, 1000,
+                     "pyinstance%d.foo", python_instance_number);
+            ++python_instance_number;
+            
+            m = bindmodule(instance_name);
+
+            /* This is expected to get bound method. */
+            method = PyObject_GetAttr(py_result, method_name);
+            
+            r = bindrule( imported_method_name, root_module() );
+
+            r->python_function = method;
+
+            result = list_new(0, newstr(instance_name));    
+
+            Py_DECREF(method_name);
+        }
+        else if (py_result == Py_None)
+        {
+            result = L0;
+        }
+        else
+        {
+            fprintf(stderr, "Non-list object returned by Python call\n");
+        }
+
+        Py_DECREF(py_result);
     }
-    /* pop this stack frame */
-    --frame->info->stack_count;
-    profile_stack = frame->caller;
+    else {
+        PyErr_Print();
+        fprintf(stderr,"Call failed\n");
+    }
+    
+    return result;
 }
 
-static void dump_profile_entry(void* p_, void* ignored)
+module_t* python_module()
 {
-    profile_info* p = (profile_info*)p_;
-    printf("%10d %10d %10d %s\n", p->cumulative, p->net, p->num_entries, p->name);
+    static module_t* python = 0;
+    if ( !python )
+        python = bindmodule("__python__");
+    return python;
 }
 
-void profile_dump()
-{
-    if ( profile_hash )
-    {
-        printf("%10s %10s %10s %s\n", "gross", "net", "# entries", "name");
-        hashenumerate( profile_hash, dump_profile_entry, 0 );
-    }
-}
+#endif
 
 /*
  * evaluate_rule() - execute a rule invocation
@@ -922,6 +946,47 @@ evaluate_rule(
     rulename = l->string;
     rule = bindrule( l->string, frame->module );
 
+#ifdef HAVE_PYTHON
+    if (rule->python_function)
+    {
+        /* The below messing with modules is due to the
+           way modules are implemented in jam.  
+           Suppose we're in module M1 now.  The global
+           variable map actually holds 'M1' variables,
+           and M1->variables hold global variables.
+
+           If we call Python right away, and then Python
+           call back Jam, and jam does:
+
+              module M1 {  }
+
+           then jam will try to swap current global 
+           variables with M1->variables. The result will
+           be that global variables map will hold
+           global variables, and any variables settings
+           we do will go to global module, not M1.
+
+           By restoring basic state, where global
+           variable map hold global variable, we make
+           sure any fugure 'module M1' will work OK.  */
+           
+        LIST *result;
+        module_t *m = python_module();
+
+        frame->module = m;
+        
+        exit_module( prev_module );
+        enter_module( m );
+
+        result = call_python_function(rule, frame);
+
+        exit_module( m );
+        enter_module ( prev_module );
+
+        return result;
+    }
+#endif
+
     /* drop the rule name */
     l = list_pop_front( l );
 
@@ -987,12 +1052,31 @@ evaluate_rule(
 
         /* The action is associated with this instance of this rule */
 
-        action = (ACTION *)malloc( sizeof( ACTION ) );
+        action = (ACTION *)BJAM_MALLOC( sizeof( ACTION ) );
         memset( (char *)action, '\0', sizeof( *action ) );
 
         action->rule = rule;
         action->targets = targetlist( (TARGETS *)0, lol_get( frame->args, 0 ) );
         action->sources = targetlist( (TARGETS *)0, lol_get( frame->args, 1 ) );
+        
+        /*  Make targets[1,N-1] depend on targets[0], to describe the multply
+            generated targets for the rule. Do it with includes, to reflect
+            non-build dependency. */
+        if ( action->targets )
+        {
+            TARGET * t0 = action->targets->target;
+            for ( t = action->targets->next; t; t = t->next )
+            {
+                TARGET * tn = t->target;
+                if ( !tn->includes )
+                {
+                    tn->includes = copytarget( tn );
+                    tn->includes->original_target = tn;
+                }
+                tn = tn->includes;
+                tn->depends = targetentry( tn->depends, t0 );
+            }
+        }
 
         /* Append this action to the actions of each target */
 
@@ -1049,7 +1133,9 @@ LIST *call_rule( char *rulename, FRAME* caller_frame, ...)
     FRAME       inner[1];
     frame_init( inner );
     inner->prev = caller_frame;
-    inner->module = caller_frame->module;
+    inner->prev_user = caller_frame->module->user_module ? 
+        caller_frame : caller_frame->prev_user;
+    inner->module = caller_frame->module;    
     inner->procedure = 0;
 
     va_start(va, caller_frame);    
@@ -1090,6 +1176,26 @@ compile_rules(
     return parse_evaluate( parse, frame );
 }
 
+/*
+ * assign_var_mode() - convert ASSIGN_XXX compilation flag into
+ *                     corresponding VAR_XXX variable set flag.
+ */
+static int assign_var_mode(int parsenum, const char **tracetext)
+{
+    const char *trace;
+    int setflag;
+    switch( parsenum )
+    {
+        case ASSIGN_SET:     setflag = VAR_SET; trace = "="; break;
+        case ASSIGN_APPEND:  setflag = VAR_APPEND; trace = "+="; break;
+        case ASSIGN_DEFAULT: setflag = VAR_DEFAULT; trace = "?="; break;
+        default:             setflag = VAR_SET; trace = ""; break;
+    }
+    if (tracetext)
+        *tracetext = trace ;
+    return setflag;
+}
+
 /*
  * compile_set() - compile the "set variable" statement
  *
@@ -1106,16 +1212,8 @@ compile_set(
     LIST    *nt = parse_evaluate( parse->left, frame );
     LIST    *ns = parse_evaluate( parse->right, frame );
     LIST    *l;
-    int     setflag;
-    char    *trace;
-
-    switch( parse->num )
-    {
-    case ASSIGN_SET:    setflag = VAR_SET; trace = "="; break;
-    case ASSIGN_APPEND: setflag = VAR_APPEND; trace = "+="; break;
-    case ASSIGN_DEFAULT:    setflag = VAR_DEFAULT; trace = "?="; break;
-    default:        setflag = VAR_SET; trace = ""; break;
-    }
+    const char *trace;
+    int     setflag = assign_var_mode( parse->num, &trace );
 
     if( DEBUG_COMPILE )
     {
@@ -1207,7 +1305,8 @@ compile_settings(
     LIST    *ns = parse_evaluate( parse->third, frame );
     LIST    *targets = parse_evaluate( parse->right, frame );
     LIST    *ts;
-    int append = parse->num == ASSIGN_APPEND;
+    const char *trace;
+    int     setflag = assign_var_mode( parse->num, &trace );
 
     if( DEBUG_COMPILE )
     {
@@ -1215,7 +1314,7 @@ compile_settings(
         list_print( nt );
         printf( " on " );
         list_print( targets );
-        printf( " %s ", append ? "+=" : "=" );
+        printf( " %s ", trace );
         list_print( ns );
         printf( "\n" );
     }
@@ -1230,7 +1329,7 @@ compile_settings(
         LIST    *l;
 
         for( l = nt; l; l = list_next( l ) )
-        t->settings = addsettings( t->settings, append, 
+        t->settings = addsettings( t->settings, setflag, 
                 l->string, list_copy( (LIST*)0, ns ) );
     }
 
